import codecs
import ctypes

from enum import IntEnum

def _error_restype(code):
    code = ctypes.c_uint32(code).value
    
    if code != _ErrorCode.FT_OK:
        raise KinesisError(code)

    return _ErrorCode(code)


class KinesisError(Exception):
    def __init__(self, code: int):
        try:
            self.code    = _ErrorCode(code)
            self.name    = self.code.name
            self.message = self.code.message
        except ValueError:
            self.code    = code
            self.name    = "UnknownError"
            self.message = "Unknown Kinesis SDK Error."

        super().__init__(f"[{self.name}] (0x{int(code):08X}) {self.message}")

class _ErrorCode(IntEnum):
    FT_OK = 0x00

    # FTDI and Communication errors
    FT_InvalidHandle         = 0x01
    FT_DeviceNotFound        = 0x02
    FT_DeviceNotOpened       = 0x03
    FT_IOError               = 0x04
    FT_InsufficientResources = 0x05
    FT_InvalidParameter      = 0x06
    FT_DeviceNotPresent      = 0x07
    FT_IncorrectDevice       = 0x08

    # Device libraries errors
    FT_NoDLLLoaded          = 0x10
    FT_NoFunctionsAvailable = 0x11
    FT_FunctionNotAvailable = 0x12
    FT_BadFunctionPointer   = 0x13
    FT_GenericFunctionFail  = 0x14
    FT_SpecificFunctionFail = 0x15

    # General DLL control errors
    TL_ALREADY_OPEN           = 0x20
    TL_NO_RESPONSE            = 0x21
    TL_NOT_IMPLEMENTED        = 0x22
    TL_FAULT_REPORTED         = 0x23
    TL_INVALID_OPERATION      = 0x24
    TL_DISCONNECTING          = 0x28
    TL_FIRMWARE_BUG           = 0x29
    TL_INITIALIZATION_FAILURE = 0x2A
    TL_INVALID_CHANNEL        = 0x2B

    # Motor specific errors
    TL_UNHOMED                    = 0x25
    TL_INVALID_POSITION           = 0x26
    TL_INVALID_VELOCITY_PARAMETER = 0x27
    TL_CANNOT_HOME_DEVICE         = 0x2C
    TL_JOG_CONTINOUS_MODE         = 0x2D
    TL_NO_MOTOR_INFO              = 0x2E
    TL_CMD_TEMP_UNAVAILABLE       = 0x2F

    @property
    def message(self) -> str:
        return {
            # FTDI and Communication errors
            _ErrorCode.FT_InvalidHandle         : "The FTDI functions have not been initialized"            ,
            _ErrorCode.FT_DeviceNotFound        : "The device could not be found"                           ,
            _ErrorCode.FT_DeviceNotOpened       : "The device must be opened before it can be accessed"     ,
            _ErrorCode.FT_IOError               : "An I/O Error has occured in the FTDI chip"               ,
            _ErrorCode.FT_InsufficientResources : "There are Insufficient resources to run this application",
            _ErrorCode.FT_InvalidParameter      : "An invalid parameter has been supplied to the device"    ,
            _ErrorCode.FT_DeviceNotPresent      : "The device is no longer present"                         ,
            _ErrorCode.FT_IncorrectDevice       : "The device detected does not match that expected."       ,

            # Device libraries errors
            _ErrorCode.FT_NoDLLLoaded          : "The library for this device could not be found",
            _ErrorCode.FT_NoFunctionsAvailable : "No functions available for this device"        ,
            _ErrorCode.FT_FunctionNotAvailable : "The function is not available for this device" ,
            _ErrorCode.FT_BadFunctionPointer   : "Bad function pointer detected"                 ,
            _ErrorCode.FT_GenericFunctionFail  : "The function failed to complete succesfully"   ,
            _ErrorCode.FT_SpecificFunctionFail : "The function failed to complete succesfully"   ,

            # General DLL control errors
            _ErrorCode.TL_ALREADY_OPEN           : "Attempt to open a device that was already open"                        ,
            _ErrorCode.TL_NO_RESPONSE            : "The device has stopped responding"                                     ,
            _ErrorCode.TL_NOT_IMPLEMENTED        : "The function has not been implemented"                                 ,
            _ErrorCode.TL_FAULT_REPORTED         : "The device has reported a fault"                                       ,
            _ErrorCode.TL_INVALID_OPERATION      : "The function could not be completed at this time"                      ,
            _ErrorCode.TL_DISCONNECTING          : "The function could not be completed because the device is disconnected",
            _ErrorCode.TL_FIRMWARE_BUG           : "The firmware has thrown an error"                                      ,
            _ErrorCode.TL_INITIALIZATION_FAILURE : "The device failed to initialize"                                       ,
            _ErrorCode.TL_INVALID_CHANNEL        : "An invalid channel address was supplied"                               ,

            # Motor specific errors
            _ErrorCode.TL_UNHOMED                    : "The device cannot perform this function until it has been homed"                   ,
            _ErrorCode.TL_INVALID_POSITION           : "The function cannot be performed as it would result in an illegal position"        ,
            _ErrorCode.TL_INVALID_VELOCITY_PARAMETER : "An invalid velocity parameter was supplied. The velocity must be greater than zero",
            _ErrorCode.TL_CANNOT_HOME_DEVICE         : "The device does not support homing. Check the limit switch parameters are correct" ,
            _ErrorCode.TL_JOG_CONTINOUS_MODE         : "An invalid jog mode was supplied for the jog function"                             ,
            _ErrorCode.TL_NO_MOTOR_INFO              : "There is no Motor Parameters available to convert Real World Units"                ,
            _ErrorCode.TL_CMD_TEMP_UNAVAILABLE       : "Command temporarily unavailable. Device may be busy"
        
        }.get(self, "Unknown Kinesis SDK error.")


from .core._kcube_solenoid import _OperatingMode, _OperatingState, _SolenoidState
from .core._kcube_solenoid import _buildDeviceList, _getDeviceListExt, _open, _close, _getOperatingMode, \
                                  _setOperatingMode, _getSolenoidState, _getOperatingState, _setOperatingState

class KCubeSolenoid:

    @staticmethod
    def listDevices() -> list[str]:
        _buildDeviceList()
        return _getDeviceListExt()

    @classmethod
    def fromSerialNo(cls, serialNo: str):
        _buildDeviceList()
        deviceList = _getDeviceListExt()

        # Check if serial number is valid
        serialNo = str(serialNo)
        if serialNo not in deviceList:
            raise ValueError(f"Serial number {serialNo} not found")

        return cls(serialNo)

    @classmethod
    def fromIndex(cls, index: int):
        # Check if index is valid
        _buildDeviceList()
        deviceList = _getDeviceListExt()
        if index >= len(deviceList):
            raise IndexError(f"Solenoid index {index} out of range (available solenoids: {len(deviceList)})")

        serialNo = deviceList[index]
        return cls(serialNo)

    def __init__(self, serialNo: str):

        _buildDeviceList()
        deviceList = _getDeviceListExt()
        
        # Check if serial number is valid
        serialNo = str(serialNo)
        if serialNo not in deviceList:
            raise ValueError(f"Serial number {serialNo} not found")

        self._index    = deviceList.index(serialNo)
        self._serialNo = serialNo

        # Opening the instance
        _open(self._serialNo)
        self._isClosed = False

        # Initialize state and mode. For safety it will
        # always originally be set to "Manual / Closed"
        self.operatingMode  = "manual"
        self.operatingState = "inactive" # ie., closed

    def __enter__(self):
        return self


    @property
    def operatingMode(self) -> str:
        return _getOperatingMode(self._serialNo).label

    @operatingMode.setter
    def operatingMode(self, operatingMode: str):
        _setOperatingMode(self._serialNo, _OperatingMode(operatingMode))

    @property
    def operatingState(self) -> str:
        return _getOperatingState(self._serialNo).label

    @operatingState.setter
    def operatingState(self, operatingState: str):
        _setOperatingState(self._serialNo, _OperatingState(operatingState))

    def openSolenoid(self):
        # Force the solenoid to be opened until new update
        self.operatingMode  = "manual"
        self.operatingState = "active" # ie., opened

    def closeSolenoid(self):
        # Force the solenoid to be closed until new update
        self.operatingMode  = "manual"
        self.operatingState = "inactive" # ie., closed

    # Pythonic aliases
    open  = openSolenoid
    close = closeSolenoid


    def _closeInstance(self):
        if not self._isClosed:
            _close(self._serialNo)
            self._isClosed = True

    def __del__(self):
        self._closeInstance()

    def __exit__(self, exceptionType, exceptionValue, traceback):
        self._closeInstance()


import ctypes
import enum
import platform
import os

from ._errors import _error_restype


# Cosing and reading correct dll
system = platform.system()
arch   = platform.architecture()[0]
if system == "Windows":
    if arch == '64bit':
        dllPath = os.path.join(os.path.dirname(__file__), '../dependencies', system, 'x64', 'Thorlabs.MotionControl.KCube.Solenoid.dll')
    else:
         raise ValueError(f"Unsupported architecture: {arch}. Currently only 64bit is supported.")
else:
    raise ValueError(f"Unsupported system: {system}. Currently only Windows is supported.")
lib = ctypes.cdll.LoadLibrary(dllPath)


# Number of enum and structs binded: 3 / 12

# typedef enum SC_OperatingModes : byte
class _OperatingMode(enum.IntEnum):
    _Manual    = 0x01
    _Single    = 0x02
    _Auto      = 0x03
    _Triggered = 0x04

    @property
    def label(self) -> str:
        mapping = {
            self._Manual   : "Manual",
            self._Single   : "Single",
            self._Auto     : "Auto",
            self._Triggered: "Triggered",
        }
        return mapping.get(self, "Unknown")

    @classmethod
    def _missing_(cls, value: str):
        value   = value.strip().lower()
        mapping = {
            "manual"   : cls._Manual,
            "single"   : cls._Single,
            "auto"     : cls._Auto,
            "triggered": cls._Triggered,
        }
        if value in mapping: return mapping[value]

        raise ValueError(f"{value} is not a valid {cls.__name__}")

# typedef enum SC_OperatingStates : byte
class _OperatingState(enum.IntEnum):
    _Active   = 0x01
    _Inactive = 0x02

    @property
    def label(self) -> str:
        mapping = {
            self._Active  : "Active",
            self._Inactive: "Inactive",
        }
        return mapping.get(self, "Unknown")

    @classmethod
    def _missing_(cls, value: str):
        value   = value.strip().lower()
        mapping = {
            "active"  : cls._Active,
            "inactive": cls._Inactive,
        }
        if value in mapping: return mapping[value]

        raise ValueError(f"{value} is not a valid {cls.__name__}")

# typedef enum SC_SolenoidStates : byte
class _SolenoidState(enum.IntEnum):
    _Opened = 0x01
    _Closed = 0x02

    @property
    def label(self) -> str:
        mapping = {
            self._Opened: "Opened",
            self._Closed: "Closed",
        }
        return mapping.get(self, "Unknown")

    @classmethod
    def _missing_(cls, value: str):
        value   = value.strip().lower()
        mapping = {
            "opened": cls._Opened,
            "closed": cls._Closed,
        }
        if value in mapping: return mapping[value]

        raise ValueError(f"{value} is not a valid {cls.__name__}")

# Number of functions binded: 10 / 68


# KCUBESOLENOID_API short __cdecl TLI_BuildDeviceList(void)
lib.TLI_BuildDeviceList.restype  =  _error_restype
lib.TLI_BuildDeviceList.argtypes = []
def _buildDeviceList() -> None:
    lib.TLI_BuildDeviceList()

# KCUBESOLENOID_API short __cdecl TLI_GetDeviceListSize()
# ------------------------- TO-DO -----------------------

# KCUBESOLENOID_API short __cdecl TLI_GetDeviceList(SAFEARRAY** stringsReceiver)
# This function should not be used in the Python binding. It should be preferred
# the following : TLI_GetDeviceListExt(char *receiveBuffer, DWORD sizeOfBuffer).

# KCUBESOLENOID_API short __cdecl TLI_GetDeviceListByType(SAFEARRAY** stringsReceiver, int typeID)
# ------------------------- TO-DO ----------------------------------------------------------------

# KCUBESOLENOID_API short __cdecl TLI_GetDeviceListByTypes(SAFEARRAY** stringsReceiver, int * typeIDs, int length)
# ------------------------- TO-DO --------------------------------------------------------------------------------

# KCUBESOLENOID_API short __cdecl TLI_GetDeviceListExt(char *receiveBuffer, DWORD sizeOfBuffer)
lib.TLI_GetDeviceListExt.restype  =  _error_restype
lib.TLI_GetDeviceListExt.argtypes = [ctypes.c_char_p, ctypes.c_uint]
def _getDeviceListExt() -> list[str]:
    bufferSize = 256
    buffer     = ctypes.create_string_buffer(bufferSize)

    lib.TLI_GetDeviceListExt(buffer, bufferSize)
    devices = buffer.value.decode('utf-8')

    return [device for device in devices.split(',') if device]

# KCUBESOLENOID_API short __cdecl TLI_GetDeviceListByTypeExt(char *receiveBuffer, DWORD sizeOfBuffer, int typeID)
# ------------------------- TO-DO -------------------------------------------------------------------------------

# KCUBESOLENOID_API short __cdecl TLI_GetDeviceListByTypesExt(char *receiveBuffer, DWORD sizeOfBuffer, int * typeIDs, int length)
# ------------------------- TO-DO -----------------------------------------------------------------------------------------------

# KCUBESOLENOID_API short __cdecl TLI_GetDeviceInfo(char const * serialNo, TLI_DeviceInfo *info)
# ------------------------- TO-DO --------------------------------------------------------------

# KCUBESOLENOID_API void __cdecl TLI_InitializeSimulations()
# ------------------------- TO-DO --------------------------

# KCUBESOLENOID_API void __cdecl TLI_UninitializeSimulations()
# ------------------------- TO-DO ----------------------------

# KCUBESOLENOID_API short __cdecl SC_Open(char const * serialNo)
lib.SC_Open.restype  =  _error_restype
lib.SC_Open.argtypes = [ctypes.c_char_p]
def _open(serialNo: str) -> None:
    lib.SC_Open(serialNo.encode('utf-8'))

# KCUBESOLENOID_API void __cdecl SC_Close(char const * serialNo)
lib.SC_Close.restype  =  _error_restype
lib.SC_Close.argtypes = [ctypes.c_char_p]
def _close(serialNo: str) -> None:
    lib.SC_Close(serialNo.encode('utf-8'))

# KCUBESOLENOID_API bool __cdecl SC_CheckConnection(char const * serialNo)
# ------------------------- TO-DO ----------------------------------------

# KCUBESOLENOID_API void __cdecl SC_Identify(char const * serialNo)
# ------------------------- TO-DO ---------------------------------

# 27 functions are still missing before the next one

# KCUBESOLENOID_API SC_OperatingModes __cdecl SC_GetOperatingMode(char const * serialNo)
lib.SC_GetOperatingMode.restype  =  ctypes.c_ubyte # Prefer unsigned rather than c_byte
lib.SC_GetOperatingMode.argtypes = [ctypes.c_char_p]
def _getOperatingMode(serialNo: str) -> _OperatingMode:
    mode = lib.SC_GetOperatingMode(serialNo.encode("utf-8"))

    try:               return _OperatingMode(mode)
    except ValueError: raise   RuntimeError(f"Unknown operating mode: {mode}")

# KCUBESOLENOID_API short __cdecl SC_SetOperatingMode(char const * serialNo, SC_OperatingModes mode)
lib.SC_SetOperatingMode.restype  =  _error_restype
lib.SC_SetOperatingMode.argtypes = [ctypes.c_char_p, ctypes.c_ubyte] # Prefer c_ubyte than c_byte
def _setOperatingMode(serialNo: str, operatingMode: _OperatingMode) -> None:
    lib.SC_SetOperatingMode(serialNo.encode("utf-8"), operatingMode)

# KCUBESOLENOID_API SC_SolenoidStates __cdecl SC_GetSolenoidState(char const * serialNo)
lib.SC_GetSolenoidState.restype  =  ctypes.c_ubyte # Prefer unsigned rather than c_byte
lib.SC_GetSolenoidState.argtypes = [ctypes.c_char_p]
def _getSolenoidState(serialNo: str) -> _SolenoidState:
    state = lib.SC_GetSolenoidState(serialNo.encode("utf-8"))

    try:               return _SolenoidState(state)
    except ValueError: raise   RuntimeError(f"Unknown solenoid state: {state}")

# KCUBESOLENOID_API short __cdecl SC_RequestOperatingState(char const * serialNo)
# ------------------------- TO-DO -----------------------------------------------

# KCUBESOLENOID_API SC_OperatingStates __cdecl SC_GetOperatingState(char const * serialNo)
lib.SC_GetOperatingState.restype  =  ctypes.c_ubyte # Prefer unsigned rather than c_byte
lib.SC_GetOperatingState.argtypes = [ctypes.c_char_p]
def _getOperatingState(serialNo: str) -> _OperatingState:
    state = lib.SC_GetOperatingState(serialNo.encode("utf-8"))

    try:               return _OperatingState(state)
    except ValueError: raise   RuntimeError(f"Unknown operating state: {state}")

# KCUBESOLENOID_API short __cdecl SC_SetOperatingState(char const * serialNo, SC_OperatingStates state)
lib.SC_SetOperatingState.restype  =  _error_restype
lib.SC_SetOperatingState.argtypes = [ctypes.c_char_p, ctypes.c_ubyte] # Prefer c_ubyte than c_byte
def _setOperatingState(serialNo: str, operatingState: _OperatingState) -> None:
    lib.SC_SetOperatingState(serialNo.encode("utf-8"), operatingState)

# 20 functions are still missing before the end of the file


# High level convenience class
from .kcube_solenoid import KCubeSolenoid

from importlib.metadata import version, PackageNotFoundError
try:
    __version__ = version("pykinesis")
except PackageNotFoundError:
    __version__ = "?.?.?"